import copy

import numpy as np

from scipy.spatial import distance


''' Swarm Cluster for PSO calculations on HiC data to find optimal distances for xyz points of beads
        This creates a cluster of particles, each particle is represented by a list of xyz coordinates, 
        a point of each particle is the individual coordinate (x,y, or z) of each bead. A different velocity is applied
        to each individual coordinate (x, y, or z).

    Swarms are in a cluster format as this allows python to do vector math on the swarm array which is MANY times more efficient
    than doing individual objects of swarms
    
    Developer: 
    David Vadnais dvadnais@uccs.edu
    DMichael Middleton mmiddlet@uccs.edu 12/4/2020

'''

class Swarm:
    id = 0

    def __init__(self, ref, pointCount, randVal = 0.5, swarmSize = 15, zeroInd=None, swarmComb = None , lossFunc = 0):
        '''
        ref: distance matrix from HiC data \n
        pointCount: Number of beads \n
        randVal: random value to calculate initial x,y,z from \n
        swarmSize: number of particles in cluster \n
        zeroInd: Used to delete in distance calculation \n
        swarmComb: used to combine multiple swarms if doing multiple passes \n
        lossFunctionChoice: chosenLossFunc
        '''

        Swarm.id += 1
        self.id = Swarm.id # ID of swarm if using multiprocessing with multiple swarm clusters

        self.pc = pointCount # pointCount

        # Random Values to get initial xyz coordinates from
        self.randMax = randVal
        self.randMin = -randVal
        self.lossFunc = lossFunc

        self.gBest = None # Global best position

        self.ref = ref # Reference distance matrix
        self.zeroInd = zeroInd

        # Creates a list of particles where each particle is a list of xyz coordinates of size pointCount
        tempList = []
        if swarmComb is None:
            for i in range(swarmSize):
                tempList.append(self.Rand_Cur())
        else:
            for swarm in swarmComb:
                tempList.append(swarm.gBest[0])

            for i in range(swarmSize-len(swarmComb)):
                cutSize = np.random.randint(1, (int)(self.pc/2))
                randCopy = np.random.choice(swarmComb).gBest[0]
                tempPos, mask = self.Rand_shift(randCopy, cutSize, 0.1)
                tempList.append(tempPos)

        self.pos = np.asarray(tempList) # Turns the list into a 3D matrix for vector operations
        self.posBest = copy.copy(self.pos) # Best position of each individual particle
        self.costBest = np.full((self.pos.shape[0],1), np.inf) # Best cost of each individual particle

        self.vel = np.zeros((self.pos.shape)) # Velocity of each particle 
        self.cost = np.full((self.pos.shape[0],1), np.inf) # Cost of each particle

        self.dist = np.zeros((self.pos.shape[0],self.ref.shape[0])) # Distance of each particle

        self.locOpCount = np.zeros((self.pos.shape[0],1)) # Used to check if a particle is not changing after a certain amount of itterations

        self.Cost() # Gets first cost calculations

    def Rand_Cur(self):
        ''' Gets a array of xyz coordinates of size pointCount,3 between a minimum and maximum value '''
        return np.random.uniform(self.randMin,self.randMax, size=(self.pc, 3))

    def Rand_shift(self, copyPos, cutSize, threshold=0.1):
        '''
        Performs a random shift of all the xyz positions in a passed in position array. \n

        Args: 
            copyPos: position array to shift  
            cutSize: Number of values to shift 
            threshold: range of the random shift applied to the positions; values will be uniformly sampled from [-threshold, threshold]
        '''

        temp = copy.copy(copyPos)

        # random boolean mask for which values will be changed
        falseVal = np.zeros(cutSize).astype(bool)
        trueVal = np.ones(temp.shape[0]-cutSize).astype(bool)

        randBool = np.append(falseVal, trueVal)
        np.random.shuffle(randBool)

        mask =  np.tile(randBool.transpose(), (temp.shape[1], 1)).transpose()

        # random matrix the same shape of your data
        r = np.random.uniform(-threshold, threshold, size=temp.shape)

        temp[mask] += r[mask]

        return temp, mask

    def Update_Cost(self, newCost):
        ''' Updates the cost of each particle in cluster '''
        tmpMsk = newCost > self.cost

        self.locOpCount[tmpMsk] += np.ones(self.locOpCount.shape)[tmpMsk]

        self.cost = newCost

        tmpMsk = self.cost < self.costBest
        tmpMsk = tmpMsk.reshape(tmpMsk.shape[0])

        self.posBest[tmpMsk] = copy.copy(self.pos)[tmpMsk]
        self.costBest[tmpMsk] = self.cost[tmpMsk]

        currentBest = np.argmin(self.cost)

        if (self.gBest is None) or (self.cost[currentBest] < self.gBest[1]):
            self.gBest = (copy.copy(self.pos[currentBest]),copy.copy(self.cost[currentBest][0]), copy.copy(self.dist[currentBest]))

    
    def Update_Pos(self, itt):
        '''
        Updates the position of each particle \n
        X(t+1) = X(t)+V(t+1) \n
        
        Args:
            itt: the current iteration of program
        '''

        # Gets a random cutize
        cutSize = np.random.randint(1, (int)(self.pc-1))

        # Adjust threshold depending on iteration to gradually adjust
        if (itt > 500):
            thresh = (1/itt)*100
        else:
            thresh = 1

        # Gets a boolean array of values that checks if a certain number of particles has changed in the last iteration
        # This certain number is calculated in Calc_Const and is based on the number of iterations
        # The boolean array that is calculated is then used to shift any particle that have been calculated to not been changing
        tmpMsk = self.locOpCount > self.Calc_Const(10000,itt,5,15)
        tmpMsk = tmpMsk.reshape(self.locOpCount.shape[0])
        changeInd = np.where(tmpMsk)[0]

        # For each particle that is being changed
        for i in changeInd:
            # If iterations are over a certain amount a full new swarm is calculated
            if(itt < 1000):
                self.pos[i] = self.Rand_Cur()
            else:
                # Otherwise, shift only the cutsize amount of particles			
                self.pos[i], mask = self.Rand_shift(self.pos[i], cutSize, thresh)
                self.vel[i][mask] = np.zeros((self.vel.shape[1], self.vel.shape[2]))[mask]

        # Changes all shifted nodes to have a velocity of zero
        self.vel[tmpMsk] = np.zeros(self.vel.shape)[tmpMsk]
        # Resets local optima counter for changed particle
        self.locOpCount[tmpMsk] = -1

        # Adjusts new positions with velocitys
        self.pos[~tmpMsk] = self.pos[~tmpMsk]+self.vel[~tmpMsk]

    def Calc_Dist(self):
        ''' Calculates the euclidean distance for each particle '''
        for i in range(self.pos.shape[0]):
            if self.zeroInd is None:
                self.dist[i] = distance.pdist(self.pos[i])
            else:
                tmpDist = distance.pdist(self.pos[i])
                tmpDist = np.delete(tmpDist, self.zeroInd)
                self.dist[i] = tmpDist

        return self.dist
    
    def Cost(self):
        ''' Calculates the cost of each particle in the swarm cluster '''
        self.Calc_Dist()
        #print("which lf?" )
        #print(self.lossFunc)
        if (self.lossFunc == 1): 
            newCost =  (1/self.pc) * np.sum( (self.dist-self.ref[:,3])**2, axis=1 ) /len(self.ref[:,3])#MSE
        elif(self.lossFunc == 2): 
            #print("loss func : RMSE" )
            newCost = newCost = np.sqrt(np.sum( (self.dist-self.ref[:,3])**2, axis=1 ))#RMSE
        elif(self.lossFunc == 3): 
            #Heuber
            delta = 0.1
            y = self.ref[:,3]
            yHat = self.dist
            newCost = np.sum(np.where(np.abs(y-yHat) < delta,.5*(y-yHat)**2 , delta*(np.abs(y-yHat)-0.5*delta)),axis=1)
            #print("ME:" ,newCost)
            #print("notME",dnewCost)
        else :
            newCost = np.sum( (self.dist-self.ref[:,3])**2, axis=1 )#SSE
        
        newCost = newCost.reshape((self.pos.shape[0],1))

        self.Update_Cost(newCost)

        return self.cost

    def Calc_Vel(self, ittMax, itt):
        '''
        Calculates the velocitys of each particle \n
        V(t+1) = weight*V(t) + conP*ranP(pBest-X(t)) + conG*ranG(gBest-x(t))
        '''

        ranP = np.random.rand(self.pos.shape[0],self.pos.shape[1],self.pos.shape[2])#local rand  variation
        ranG = np.random.rand(self.pos.shape[0],self.pos.shape[1],self.pos.shape[2])#global rand variarion

        # Way to calculate constants, not used anymore but for future reference
        #weight = self.Calc_Const(ittMax,itt,0.1,0.5)
        #conP = self.Calc_Const(ittMax,itt,0.1,0.5)
        #conG = self.Calc_Const(ittMax,itt,0.1,2)

        # Constants found to work the best
        weight = 0.5
        conP=0.3 #Local confidence
        conG=2.5 #Global confidence

        self.vel = (weight*self.vel) + (conP*ranP*(self.posBest-self.pos)) + (conG*ranG*(self.gBest[0]-self.pos))

        return self.vel 

    # 
    def Calc_Const(self, ittMax, k, W_MIN, W_MAX):
        '''
        Function that calculates a constant value

        Args:
            ittMax: maximum number of itterations
            k: current itterations
            W_MIN: constant min
            W_MAX: constant max
        '''
        if ittMax < k:
            k = ittMax
        return ( ( (W_MAX-W_MIN)*((ittMax-k)/ittMax) ) + W_MIN )



import sys
import argparse
import os
import logging

import numpy as np
from scipy import stats

from multiprocessing import Pool, cpu_count
from itertools import repeat

from ParticleChromo3D.Swarm import Swarm
import ParticleChromo3D.Helper as Helper
from ParticleChromo3D.Helper import Write_Log
from ParticleChromo3D.particle_chromo_logger import setup_logger

# email imports
import smtplib
import uuid
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.header import Header

# Initialize logger
logger = logging.getLogger(__name__)

# I should refactor these but I dont think the interest
randRange = None
swarmSize = None
ittCount = None


# Function defs
def lossFunction(tar, b, lossFunctionChoice):
    if lossFunctionChoice == 2:
        newCost = np.sqrt(np.sum((b - tar) ** 2))  # RMSE
    elif lossFunctionChoice == 0:
        newCost = np.sum((b - tar) ** 2)  # SSE
    elif lossFunctionChoice == 1:
        newCost = np.square(np.subtract(b, tar)).mean()
    elif lossFunctionChoice == 3:
        # Heuber
        alpha = 0.5
        y = tar
        yHat = b
        newCost = np.sum(
            np.where(
                np.abs(y - yHat) < alpha,
                0.5 * (y - yHat) ** 2,
                alpha * (np.abs(y - yHat) - 0.5 * alpha),
            )
        )
    else:
        newCost = np.sqrt(np.sum((b - tar) ** 2))  # RMSE

    return newCost


# Prints statistics of the current swarm
def Print_Stats(swarm, contact, pointCount, i, outFilePtr, convFact):
    pers = stats.pearsonr(swarm.gBest[2], contact[:, 3])
    spear = stats.spearmanr(swarm.gBest[2], contact[:, 3])
    spearIF = stats.spearmanr(swarm.gBest[2], contact[:, 2])

    error = np.sqrt((1 / pointCount) * np.sum((swarm.gBest[2] - contact[:, 3]) ** 2))

    logger.debug('id: ' + str(swarm.id) + 
        ' itt: ' + str(i) + 
        ' Cost: ' + str(swarm.gBest[1]) + 
        ' Pearson: ' + str(pers[0]) + 
        ' Spearmen: ' + str(spear[0]) +
        ' IFSpear: ' + str(spearIF[0]) +
        ' error: ' + str(error))


def Write_Stats(swarm, contact, outFilePtr):
    Helper.Write_Output(outFilePtr, swarm.gBest[0])


# Performs one operation and prints statistics of current swarm
def One_Move(
    ittCount,
    swarm,
    contact,
    pointCount,
    threshold,
    outFilePtr,
    convFact,
    lossFunctionChoice,
):
    saveGBestCost = float("inf")
    totTime = 0

    for i in range(ittCount):
        if (i % 1000 == 0) and (swarm.gBest is not None):
            error = lossFunction(contact[:, 3], swarm.gBest[2], lossFunctionChoice)
            Print_Stats(swarm, contact, pointCount, i, outFilePtr, convFact)

            if (np.abs(saveGBestCost - error)) >= threshold:
                saveGBestCost = error
            else:
                return i, totTime

        operation(i, swarm)

    return i


# Performs a single PSO pass: Velocity calculation, update position, get new cost
def operation(i, swarm):
    swarm.Calc_Vel(ittCount, i)
    swarm.Update_Pos(i)
    swarm.Cost()


# Optimizes single swarm
def Optimize(
    inFilePtr,
    outFilePtr,
    convFact,
    constraint,
    points,
    zeroInd,
    randRange,
    swarmSize,
    threshold,
    ittCount,
    lossFunctionChoice,
):
    dist = 1.0 / (constraint[:, 2] ** convFact)
    constraint = np.insert(constraint, 3, dist, axis=1)

    swarm = Swarm(
        constraint, len(points), randVal=randRange, swarmSize=swarmSize, zeroInd=zeroInd
    )

    ittFin = One_Move(
        ittCount,
        swarm,
        constraint,
        len(points),
        threshold,
        outFilePtr,
        convFact,
        lossFunctionChoice,
    )
    return (
        stats.pearsonr(swarm.gBest[2], constraint[:, 3])[0],
        stats.spearmanr(swarm.gBest[2], constraint[:, 3])[0],
        lossFunction(constraint[:, 3], swarm.gBest[2], lossFunctionChoice),
        ittFin,
        swarm.id,
        swarm,
    )


# Runs in paralel if passed multiple rangeSpace
def Par_Choice(inFilePtr, outFilePtr, alpha, randRange, swarmSize, threshold, ittCount, lossFunctionChoice):
    contact, points, zeroInd = Helper.Read_Data(inFilePtr, alpha)

    bestSwarm = None
    if 1 == 1:
        convStore = []
        alphas = np.array(range(int(alpha[0]), int(alpha[1]), int(alpha[2]))) / 100
        pool = Pool(processes=get_proc_count())
        swarms = pool.starmap(
            Optimize,
            zip(
                repeat(inFilePtr),
                repeat(outFilePtr),
                alphas,
                repeat(contact),
                repeat(points),
                repeat(zeroInd),
                repeat(randRange),
                repeat(swarmSize),
                repeat(threshold),
                repeat(ittCount),
                repeat(lossFunctionChoice),
            ),
        )

        pool.close()
        pool.join()

        iforapl = 0
        for swarm in swarms:
            # print(str(swarm[-1]) + ' ' + str(swarm[1]))
            convStore.append(swarm)
            if (bestSwarm is None) or (swarm[1] > bestSwarm[1]):
                bestSwarm = swarm
                swarmForPDB = swarm[len(swarm) - 1]
                bestAlpha = alphas[iforapl]
            iforapl += 1
    else:
        bestSwarm = Optimize(inFilePtr, outFilePtr, alpha)
    contact = np.insert(contact, 3, 1.0 / (contact[:, 2] ** bestAlpha), axis=1)
    logger.info(f"bestSwarm : {bestSwarm}")
    Write_Stats(swarmForPDB, contact, outFilePtr)

    return bestSwarm

def strip_file(in_file):
    fout = in_file + ".stripped"
    clean_lines = []
    f = open(in_file, "r")
    lines = f.readlines()
    for line in lines:
        res = str(" ".join(line.split()))
        clean_lines.append(res)
    f.close()

    with open(fout, "w") as f:
        f.writelines("\n".join(clean_lines))
    f.close()
    return fout


def Full_List(inputFilePtr, outFilePtr, alpha,randRange=0.1, swarmSize=5, threshold=0.000001, ittCount=30000, lossFunctionChoice=2):
    convStore = []

    convStore.append(Par_Choice(inputFilePtr, outFilePtr, alpha, randRange, swarmSize, threshold, ittCount, lossFunctionChoice))
    logger.info(
        f"pearson: {convStore[0][0]}"
        + f" spearman: {convStore[0][1]}"
        + f" rmse: {convStore[0][2]}"
    )

    # Helper.Write_List(convStore, outFilePtr)
    return convStore

def get_proc_count():
    ''' attempt to reduce thrashing.'''
    if cpu_count() <= 2:
        return cpu_count()
    else:
        return cpu_count() - 1  


if __name__ == "__main__":
    sys.setrecursionlimit(10000)
    

    rangeSpace = []  # Max scaling factor. Needs to be optimized for each specific dataset. Use two values [one, two] to multithread through a range of those two values at a interval of 5000

    # Arguments for running program
    # python3 ParticleChromo3D.py <input_data> <other_parameter>
    parser = argparse.ArgumentParser("ParticleChromo3D")
    parser.add_argument("infile", help="Matrix of contacts", type=str)
    parser.add_argument(
        "-ss",
        "--swarmSize",
        help="Number of particles in system [Default 15]",
        type=int,
        default=5,
    )
    parser.add_argument(
        "-itt",
        "--ittCount",
        help="Maximum itterations before stop [Default 30000]",
        type=int,
        default=30000,
    )
    parser.add_argument(
        "-t",
        "--threshold",
        help="Error threshold before stoping [Default 0.000001]",
        type=float,
        default=0.000001,
    )
    parser.add_argument(
        "-rr",
        "--randRange",
        help="Range of x,y,z starting coords. Random value bewtween -randRange,randRange [Default 1]",
        type=float,
        default=1.0,
    )
    parser.add_argument(
        "-o",
        "--outfile",
        help="Filename of the output pdb model  [Default ./chr.pdb]",
        type=str,
        default="./out/chr.pdb",
    )
    parser.add_argument(
        "-e",
        "--email",
        help="Email to message [Default NULL]",
        type=str,
        default="NULL",
    )
    parser.add_argument(
        "-lf",
        "--lossFunction",
        help="Email to message [Default NULL]",
        type=str,
        default="2",
    )
    parser.add_argument(
        "-ll",
        "--logLevel",
        help="Log level [Default INFO]",
        type=str,
        default="INFO",
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],  # Valid log levels
    )

    

    args = parser.parse_args()

    if args.infile:
        inFilePtr = args.infile
    if args.outfile:
        outFilePtr = args.outfile + str(uuid.uuid4())
    else:
        outFilePtr = "noIn"
    if args.swarmSize:
        swarmSize = args.swarmSize
    if args.ittCount:
        ittCount = args.ittCount
    if args.threshold:
        threshold = args.threshold
    randRange = args.randRange
    if args.email:
        emailAddr = args.email
    if args.lossFunction:
        lossFunctionChoice = int(args.lossFunction)

    setup_logger(args.logLevel)  # Set logging level
    logger.info(f"Starting ParticleChromo3D with log level: {args.logLevel}")

    if len(rangeSpace) == 0:
        rangeSpace.append(20000)

    if len(rangeSpace) > 2 and (rangeSpace[0] == rangeSpace[1]):
        rangeSpace.pop()

    logger.info(f"proccessing file : {inFilePtr}")

    fout = strip_file(inFilePtr)

    theseAlphas = np.array([0.1, 2.0, 0.1]) * 100
    theAlphas = (
        np.array(range(int(theseAlphas[0]), int(theseAlphas[1]), int(theseAlphas[2])))
        / 100
    )

    if outFilePtr == "noIn":
        outFilePtr = os.path.basename(os.path.basename(inFilePtr) + str(uuid.uuid4()))
        outFilePtr = os.path.splitext(outFilePtr)[0]
        logger.info(f"In noIn outFile is {outFilePtr}")
    outputOfSwarm = Full_List(fout, outFilePtr, theseAlphas)[0]
    logger.info(outputOfSwarm)

    bestSpearm = outputOfSwarm[1]
    bestCost = outputOfSwarm[2]
    bestAlpha = theAlphas[outputOfSwarm[4]]
    bestPearsonRHO = outputOfSwarm[0]

    logger.info(f"Input file: {inFilePtr}")
    logger.info(f"Convert factor:: {bestAlpha}")
    logger.info(f"SSE at best spearman : {bestCost}")
    logger.info(f"Best Spearman correlation Dist vs. Reconstructed Dist  : {bestSpearm}")
    logger.info(f"Best Pearson correlation Dist vs. Reconstructed Dist: {bestPearsonRHO}")
    Write_Log(
        outFilePtr + ".log", inFilePtr, bestAlpha, bestCost, bestSpearm, bestPearsonRHO
    )

    if "HOSTNAME_BE" in os.environ:
        #'<br><br>neo4j by first going to https://biomlearn.uccs.edu:5000/neo and then https://biomlearn.uccs.edu:7474/browser/')
        print(
            "<br><br>Download pdb at: http://"
            + os.environ["HOSTNAME_BE"]
            + ":5001/download?ofname="
            + outFilePtr
        )

    ############################## email section

    if "SERVICE_EMAIL_KEY" not in os.environ or "SERVICE_EMAIL" not in os.environ:
        logger.warning("Missing email properties in PS.py")
        sys.exit(0)

    gmail_pass = os.environ["SERVICE_EMAIL_KEY"]
    user = os.environ["SERVICE_EMAIL"]
    host = "smtp.gmail.com"
    port = 465

    to = emailAddr
    subject = "ParticleChromo3D"
    emailBody = (
        "\n\nAfter processing the file: "
        + inFilePtr
        + " The best Spearman correlation Dist vs. Reconstructed Dist found was : "
        + str(bestSpearm)
        + ". The Best Pearson correlation Dist vs. Reconstructed Dist was : "
        + str(bestPearsonRHO)
        + ".\n\n"
    )
    inputsEBody = (
        "The parameters for this run were set to {ifname="
        + inFilePtr
        + ", ss="
        + str(swarmSize)
        + ", itt="
        + str(ittCount)
        + ", threshold="
        + str(threshold)
        + ", randRange="
        + str(randRange)
        + "} "
    )

    body = (
        "Your ParticleChromo3D+ job completed.\n The results can be found at : "
        + ": http://biomlearn.uccs.edu:5001/download?ofname="
        + outFilePtr
        + emailBody
        + inputsEBody
    )
    filename = "run.sh"

    # locate and attach desired attachments
    fileNameForEmail = outFilePtr + ".pdb"
    f = open(fileNameForEmail, "r")
    att = MIMEText(f.read())
    att.add_header("Content-disposition", "attachment", filename=fileNameForEmail)
    f.close()

    # create message object
    message = MIMEMultipart()

    # add in header
    message["From"] = Header(user)
    message["To"] = Header(to)
    message["Subject"] = Header(subject)

    # attach message body as MIMEText
    message.attach(MIMEText(body, "plain", "utf-8"))

    message.attach(att)

    #  setup email server
    server = smtplib.SMTP_SSL(host, port)
    server.login(user, gmail_pass)

    # send email and quit server
    server.sendmail(user, to, message.as_string())
    server.quit()



import numpy as np
import csv
import sys
import argparse
from ParticleChromo3D.particle_chromo_logger import setup_logger

def squarify(mat):
    '''
    takes
    [name, bin1, bin2]

    nxn where n is number of bins
    '''
    
    maxBin = 0
    for row in mat:
        maxBin = max(maxBin, row[0]+1, row[1]+1) 
        
    squareMat = np.ones((maxBin,maxBin))
    
    for row in mat:
        squareMat[row[0]][row[1]] = row[2]
        squareMat[row[1]][row[0]] = row[2]
    
    return squareMat


def main(input_file, output, delimiter = '\t'):   
    logger = setup_logger()

    threeMat = []
    with open(input_file, 'r') as f:
        reader = csv.reader(f, delimiter=delimiter)
        for index, row in enumerate(reader):
            if len(row) != 3:
                logger.warning(f"skipped line {index} with content : {row}")
                continue

            row[0] = row[0].strip() # remove spaces

            logger.debug(f"row output : {row}")
            newRow = [
                int(row[0]), 
                int(row[1]), 
                float(row[2])
            ]
            threeMat.append(newRow)


    squareMat = squarify(threeMat)

    with open(output, 'w', newline='') as tsvfile:
        writer = csv.writer(tsvfile, delimiter='\t', lineterminator='\n')
        for row in squareMat:
            writer.writerow(row)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Transform VCM file into a square matrix.")
    parser.add_argument("input_file", help="Path to the input file (TSV format)")
    parser.add_argument("-o", "--output", default="convert.out", help="Output file name (default: convert.out)")
    parser.add_argument("-d", "--delimiter", default="\t", help="Delimiter used in the input file (default: tab)")
    
    args = parser.parse_args()

    main(args.input_file,  output=args.output, delimiter=args.delimiter)
    sys.exit(0)


from neo4j import __version__ as neo4j_version
from neo4j import GraphDatabase

from ParticleChromo3D.particle_chromo_logger import setup_logger

logger = setup_logger()

class Neo4jConnection:
    
    def __init__(self, uri, user, pwd):
        logger.debug(f"neo4j version : {neo4j_version}")
        self.__uri = uri
        self.__user = user
        self.__pwd = pwd
        self.__driver = None

    def connect(self):
        try:
            self.__driver = GraphDatabase.driver(self.__uri, auth=(self.__user, self.__pwd))
        except Exception as e:
            logger.error(f"Failed to create the driver: {e}")
            raise 
        
    def close(self):
        if self.__driver is not None:
            self.__driver.close()
        
    def query(self, query, db=None):
        assert self.__driver is not None, "Driver not initialized!"
        session = None
        response = None
        try: 
            session = self.__driver.session(database=db) if db is not None else self.__driver.session() 
            response = list(session.run(query))
        except Exception as e:
            logger.error(f"Query failed: {e}")
        finally: 
            if session is not None:
                session.close()
        return response
        
    def run(self):
        query_string = '''
        // Delete all
        MATCH(n)
        DETACH DELETE n;
        '''
        self.query(query_string, db='neo4j')

        file = "out/chr.pdb"

        # Using readlines()
        file1 = open(file, 'r')
        Lines = file1.readlines()
        
        # Strips the newline character
        notFirstNode = False
        lastAtom = ""
        for line in Lines:
            
            if len(line) > 4 and line[3] == 'M':
                noSpace = " ".join(line.split())
                arr = [ x.strip() for x in noSpace.strip('[]').split(' ') ]
                
                query_string = \
                    "CREATE (node:"+arr[0] \
                    +"{x: " + arr[5] +\
                    ", y: " + arr[6] +\
                    ", z: " + arr[7] +\
                    ", pdbID: " + arr[1] +\
                    ", occupancy: " + arr[8] +\
                    ", temperatureFactor: " + arr[9] +\
                    "})"
                self.query(query_string, db='neo4j')
                
                if(notFirstNode):
                    query_string = \
                        "MATCH (a:ATOM), (b:ATOM)" + \
                        " WHERE a.pdbID = " + lastAtom +\
                        " AND b.pdbID = " + arr[1] + \
                        " CREATE (a)-[r: Bonded_To]->(b)"+\
                        " RETURN a,b"
                        
                    self.query(query_string, db='neo4j')
                        
                
                lastAtom = arr[1]
                notFirstNode = True
                logger.info(arr)



def __main__():
    connection = Neo4jConnection(uri="bolt://davidvadnais.com:7687", user="neo4j", pwd="DUMMYPASSWORD")
    connection.connect()
    connection.run()
    connection.close()




from Bio import PDB
from scipy.spatial import procrustes

from collections.abc import Iterable
from pathlib import Path


import numpy as np
import copy

# Reads in a matrix of contacts
def Read_Matrix_To_List(filePtr):
    contact = np.genfromtxt(filePtr, delimiter=' ')
    #delete zero contacts
    contact = contact[~np.all(contact == 0, axis=0)]
    idx = np.argwhere(np.all(contact[..., :] == 0, axis=0))
    contact = np.delete(contact, idx, axis=1)
    
    count = 0

    zeroInd = []
    stopDupe = {}
    contactList = []
    for i in range(contact.shape[0]):
        row =  contact[i]
        for j in range(row.shape[0]):
            if (i != j) and ((i,j) not in stopDupe) and ((j,i) not in stopDupe):
                stopDupe[(i,j)] = 1
                stopDupe[(j,i)] = 1
                if (row[j] > 0):
                    contactList.append( (i,j, row[j]) )
                else:
                    zeroInd.append(count)
                count += 1

    if len(zeroInd) <= 0:
        zeroInd=None

    return np.array(contactList), zeroInd

# Reads in a list format for contacts
def Read_Data_List(filePtr):
    return np.loadtxt(filePtr), None

# Reads the data and puts it into a matrix of distances
def Read_Data(filePtr, convFactor=None):
    constraint, zeroInd = Read_Matrix_To_List(filePtr)
    #constraint, zeroInd = Read_Data_List(filePtr)

    pointList = set()
    for i in range(constraint.shape[0]):
        pointList.add(constraint[i][0])
        pointList.add(constraint[i][1])

    pointMap = {}
    for i in range(len(pointList)):
        pointMap[pointList.pop()] = i

    for i in range(constraint.shape[0]):
        constraint[i][0] = pointMap[constraint[i][0]]
        constraint[i][1] = pointMap[constraint[i][1]]


    return constraint, pointMap, zeroInd

# Calcualtes the average number of contacts
def avgCalc(constraint, convFactor):
    avgIf = constraint.mean(axis=0)[2]

    constrAvg = copy.copy(constraint[:,2])

    for i in range(constraint.shape[0]):
        constrAvg[i] = constrAvg[i]/avgIf

    #avgDist = (constrAvg**convFactor).mean()

    return constrAvg#, avgDist

# Writes xyz list to a PDB
def Write_Output(filePtr, xyz):
    xyz = Scale_Arr(xyz)
    WritePDB(xyz, filePtr+".pdb")

# Scales a range of values
def Scale_Arr(xyz, minVal=-10, maxVal=10):
    oldRange = xyz.max() - xyz.min()
    newRange = (maxVal - minVal)  
    return (((xyz - xyz.min()) * newRange) / oldRange) + minVal

# Writes a list to a file
def Write_List(listToWrite, filePtr):
    with open(filePtr, 'w') as f:
        for item in listToWrite:
            if isinstance(item, Iterable):
                for subItem in item:
                    f.write("%s " % subItem)
                f.write("\n")
            else:
                f.write("%s\n" % item)

# Performs procrustes on a pdb file
def Proc_PDB(inputPtr, comparePtr):
    xyzIn = Read_PDB(inputPtr)
    xyzComp = Read_PDB(comparePtr)
    standXYZComp, newXYZ, dis = procrustes(xyzComp,xyzIn)
    print(xyzIn)
    print(xyzComp)
    print(newXYZ)
    return standXYZComp,newXYZ

'''SOURCE: https://github.com/mbglab/EVR'''
def WritePDB(positions, pdb_file, ctype = "0"):
    '''Save the result as a .pdb file'''
    Path("./out").mkdir(exist_ok=True) # Create out dir if it doesnt exist

    o_file = open(pdb_file, "w")
    o_file.write("\n")

    col1 = "ATOM"
    col3 = "CA MET"
    col8 = "0.20 10.00"

    bin_num = len(positions)

    for i in range(1, bin_num+1):
        col2 = str(i)
        col4 = "B"+col2
        col5 = "%.3f" % positions[i-1][0]
        col6 = "%.3f" % positions[i-1][1]
        col7 = "%.3f" % positions[i-1][2]
        col2 = " "*(5 - len(col2)) + col2
        col4 = col4 + " " * (6 - len(col4))
        col5 = " " * (8 - len(col5)) + col5
        col6 = " " * (8 - len(col6)) + col6
        col7 = " " * (8 - len(col7)) + col7

        col = (col1, col2, col3, col4, col5, col6, col7,col8)
        line = "%s  %s   %s %s   %s%s%s  %s\n" % col
        o_file.write(line)
    col1 = "CONECT"
    for i in range(1, bin_num+1):
        col2 = str(i)
        j = i + 1
        if j > bin_num:
            if ctype == "1":
                continue
            #j = 1
        col3 = str(j)

        col2 = " " * (5 - len(col2)) + col2
        col3 = " " * (5 - len(col3)) + col3

        line = f"{col1}{col2}{col3}\n"
        o_file.write(line)

    o_file.write("END")
    o_file.close()

'''SOURCE: https://stackoverflow.com/questions/47825542/how-to-separately-get-the-x-y-or-z-coordinates-from-a-pdb-file'''
def Read_PDB(filePtr):
    parser = PDB.PDBParser()
    _ = PDB.PDBIO()
    struct = parser.get_structure('temp',filePtr)

    xyz = []
    for model in struct:
        for chain in model:
            for residue in chain:
                for atom in residue:
                    x,y,z = atom.get_coord()
                    xyz.append(np.array([x,y,z]))

    return np.asarray(xyz)	 

def Write_Log(outfile, inFile, bestAlpha, rmse, bestSpearmanRHO,bestPearsonRHO):
    FILE = open(outfile, 'w')
    inString = ("Input file: " + str(inFile) + "\n"+"Convert factor:: " + str(bestAlpha)+ "\n" +
        "Best cost  : "+ str(rmse)+ "\n"+"Best Spearman correlation Dist vs. Reconstructed Dist  : "+ str(bestSpearmanRHO)+ "\n" + "Best Pearson correlation Dist vs. Reconstructed Dist  : "+ str(bestPearsonRHO)) + "\n"

    FILE.write(inString)
    FILE.close()
    
    



## imports
from flask import Flask, render_template, request, send_file, redirect
import subprocess
from werkzeug.utils import secure_filename
import os

import re
import glob

# constants
regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b'

# Non flask functions
def make_tree(path):
    tree = dict(name=os.path.basename(path), children=[])
    try: 
        lst = os.listdir(path)
    except OSError as ose:
       print(ose) 
        #pass #ignore errors
    else:
        for name in lst:
            fn = os.path.join(path, name)
            if os.path.isdir(fn):
                tree['children'].append(make_tree(fn))
            else:
                tree['children'].append(dict(name=name))
    return tree

# https://www.geeksforgeeks.org/check-if-email-address-valid-or-not-in-python/
def check(email):
 
    # pass the regular expression
    # and the string into the fullmatch() method
    return re.fullmatch(regex, email)
        

## Flask
app = Flask(__name__)

@app.route("/process")
def home():
  # args
  ifname = "/apt/repo/" + request.args.get('ifname', type=str)
  ss = request.args.get('ss', type=str)
  itt = request.args.get('itt', type=str)
  threshold = request.args.get('threshold', type=str)
  randRange = request.args.get('randRange', type=str)
  lf = request.args.get('lf', type=str)
  outFile = request.args.get('outFile', type=str)
  email = request.args.get('email', type=str)
  # Custom upload
  if (ifname == "/apt/repo/uploaded_file"):
    list_of_files = glob.glob('/apt/upload/*')
    ifname = "/apt/upload/" + os.path.basename(max(list_of_files, key=os.path.getctime))

 
  # cmd
  thecmd = "python3 Ps.py -ss " + ss \
    +  " -itt " + itt \
    +  " --threshold " + threshold \
    +  " --randRange " + randRange \
    +  " -lf " + lf \
    +  " --outfile " + "out/"+outFile \
    +  " --email " + email \
    +  " " + ifname
  if (not check(email)):
    return "ERROR: Valid Email Required <br> Use back to return home"
  else:
    _ = subprocess.Popen(thecmd, shell=True, stdout=subprocess.PIPE, 
                     stderr=subprocess.PIPE)
    return "Processing has begun. You will recieve an email upon its completion."  
    #return subprocess.check_output(thecmd, shell=True) # uncomment this to debug the python

@app.route("/neo")
def neo():
  return subprocess.check_output("python3 pdbToNeo4j.py", shell=True)


@app.route('/form')
def form():
    return render_template('form.html')
 
@app.route('/upload', methods = ['POST', 'GET'])
def upload():
    if request.method == 'POST':
        f = request.files['file']
        f.save("/apt/upload/"+secure_filename(f.filename))
        return redirect("http://" + os.environ['HOSTNAME_BE'] +":8080/ParticleChromo3D/main.html", code=302)
        return "File saved successfully"

@app.route('/uploaded')
def dirtree():
    path = '/apt/upload/'
    return render_template('dirtree.html', tree=make_tree(path))

@app.route('/download')
def download_file():
    ofname = request.args.get('ofname', type=str)

    path = "/apt/" + ofname + ".pdb"
    return send_file(path, as_attachment=True)

@app.route('/convert')
def convert():
  iffname = request.args.get('filename', type=str)
  path = "//apt//upload//" + iffname
  thecmd = "python3 TransformVCM.py " + path
  _ = subprocess.Popen(thecmd, shell=True, stdout=subprocess.PIPE,
                     stderr=subprocess.PIPE)

  path2="/apt/convert.out"
  return send_file(path2, as_attachment=True)


if __name__ == "__main__":
    app.run(host='0.0.0.0',port = 5001,debug=True)



from setuptools import setup, find_packages
import os

def parse_requirements():
    """Reads requirements from config/requirements.txt"""
    with open(os.path.join(os.path.dirname(__file__), 'config', 'requirements.txt')) as f:
        return [line.strip() for line in f if line.strip() and not line.startswith('#')]

setup(
    name="ParticleChromo3D",
    version="1.0.3",
    description="3D chromosome structure reconstruction using ParticleChromo3D+",
    author="Oluwadare Lab",
    author_email="dvadnais@uccs.edu",
    url="https://github.com/OluwadareLab/ParticleChromo3D_Plus",
    packages=find_packages(),
    install_requires=parse_requirements(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    long_description=open('README.md').read(),  # Read description from README
    long_description_content_type='text/markdown',  # Or 'text/x-rst' for reStructuredText
)



import logging

def setup_logger(log_level = logging.INFO):
    logger = logging.getLogger('ParticleChromo3D')  # Create a named logger
    logger.setLevel(log_level)  # You can set this to DEBUG, WARNING, etc.
    
    # Create formatters
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # Create file handler for logging to a file
    file_handler = logging.FileHandler('ParticleChromo3d.log')
    file_handler.setFormatter(formatter)

    # Create stream handler for logging to the console
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)

    # Add handlers to the logger
    logger.addHandler(file_handler)
    logger.addHandler(stream_handler)

    return logger